from typing import Callable
from typing import Dict

from .http_error import HttpError
from .http_method import HttpMethod
from .http_request import HttpRequest
from .http_response import HttpResponse
from .middleware import Middleware
from .middleware import MiddlewareHandler
from .routing import Route
from .routing import Router
from .application import HttpApplication


class RouterMiddleware(Middleware):
    def __init__(self):
        self.methods: Dict[HttpMethod, Router] = {key: Router() for key in HttpMethod}

    @classmethod
    def from_http_application(
        cls, http_application: HttpApplication
    ) -> "RouterMiddleware":
        instance = cls()
        instance.methods = http_application.methods

        return instance

    def handle(self, request: HttpRequest, next: MiddlewareHandler) -> HttpResponse:
        try:
            route, controller = self.methods[request.method].match(
                request.path
            )  # type: Route, Callable

            request.path_parameters = route.parameters
            request.route = route
            response: HttpResponse = controller(request)

            return response
        except HttpError as error:
            return HttpResponse(status=error.status_code, body=error.http_message)


__all__ = ["RouterMiddleware"]



import re
from copy import copy
from typing import Any, Dict, Optional, Pattern, Union
from typing import Callable, List, Tuple

from .http_error import NotFoundError

_ROUTE_REGEX = r"\\\{\s*(?P<var>[a-z_][a-z0-9_-]*)\s*\\\}"
_VAR_REGEX = "[^/]+"


class Route:
    def __init__(self, route: str):
        self.route = route
        self._parameters_names: List[str] = []
        self._pattern: Pattern[str] = None  # type: ignore
        self._parameters: Dict[str, str] = {}
        self.is_wildcard: bool = "*" in route

    @property
    def pattern(self) -> Pattern[str]:
        if not self._pattern:
            self._parse()
        return self._pattern

    def _parse(self) -> None:
        def _parse_var(match):
            param = match.group(1)
            self._parameters_names.append(param)
            return f"({_VAR_REGEX})"

        pattern = re.escape(self.route)
        pattern = re.sub(_ROUTE_REGEX, _parse_var, pattern, re.I | re.M)
        pattern = re.sub(r"\\\*", ".*?", pattern)
        self._pattern = re.compile("^" + pattern + "$", re.I | re.M,)

    def match(self, uri: str) -> Union[bool, "Route"]:
        matches = self.pattern.findall(uri)
        if not matches:
            return False

        if not self._parameters_names:
            return copy(self)

        if isinstance(matches[0], tuple):
            matches = list(matches[0])

        route = copy(self)
        match_index = 0
        for value in matches:
            route._parameters[self._parameters_names[match_index]] = value
            match_index += 1

        return route

    @property
    def parameters(self):
        return self._parameters

    def __str__(self) -> str:
        return self.route

    def __bool__(self) -> bool:
        return True

    def __getitem__(self, key: str) -> str:
        return self._parameters[key]

    def __contains__(self, key: str) -> bool:
        return key in self._parameters

    def get(self, key: str, default: Optional[Any] = None):
        if key in self:
            return self[key]

        return default


class Router:
    def __init__(self):
        self._routes: List[Tuple[Route, Callable]] = []

    def append(self, route: Route, handler: Callable) -> None:
        assert isinstance(route, Route), "Passed route must be instance of Route"

        self._routes.append((route, handler))
        self._routes.sort(key=lambda r: r[0].is_wildcard)

    def match(self, uri: str) -> Tuple[Route, Callable]:
        for route in self._routes:
            if route[0].match(uri):
                return route

        raise NotFoundError(f"Could not match any resource matching {uri} uri")


__all__ = ["Route", "Router"]



import base64
import os
from io import BytesIO
from typing import Any
from typing import Callable
from typing import Dict
from urllib.parse import quote_plus

from .http_error import HttpError
from .http_headers import HttpHeaders
from .http_query_string import HttpQueryString
from .http_request import HttpRequest
from .http_response import HttpResponse
from .http_status import HttpStatus
from .middleware import MiddlewareHandler
from .middleware import MiddlewarePipeline
from .routing import Route

IS_SERVERLESS_ENVIRONMENT = bool(
    os.environ.get("AWS_LAMBDA_FUNCTION_VERSION")
    or os.environ.get("LAMBDA_RUNTIME_DIR")
    or os.environ.get("AWS_LAMBDA_FUNCTION_NAME")
)

TEXT_MIME_TYPES = [
    "application/json",
    "application/javascript",
    "application/xml",
    "application/vnd.api+json",
    "image/svg+xml",
]


def is_http_api_lambda(event: Dict[str, Any]) -> bool:
    if event.get("version") and event["version"] == "2.0":
        return True

    return False


def create_http_request_from_serverless_event(
    event: Dict[str, Any], context: Dict[str, Any]
) -> HttpRequest:
    is_http_api = is_http_api_lambda(event)

    if is_http_api:
        request = create_http_request_from_serverless_http_api(event, context)
    else:
        request = create_http_request_from_serverless_rest_api(event, context)

    return request


def create_http_request_from_serverless_http_api(
    event: Dict[str, Any], context: Dict[str, Any]
) -> HttpRequest:
    body = get_normalised_body_from_serverless(event)
    request_context = event.get("requestContext", {})
    http_context = request_context.get("http", {})

    headers = get_normalised_headers_from_serverless(event)
    headers["Cookie"] = "; ".join(event.get("cookies", []))
    headers["Content-Length"] = str(body.getbuffer().nbytes)

    request = HttpRequest(
        method=http_context["method"],
        path=http_context["path"],
        body=body,
        query_string=HttpQueryString(event.get("rawQueryString", "")),
        headers=HttpHeaders(headers),
    )
    request.path_parameters = event.get("pathParameters", {})

    request.attributes["aws_context"] = context
    request.attributes["aws_event"] = event

    return request


def create_http_request_from_serverless_rest_api(
    event: Dict[str, Any], context: Dict[str, Any]
) -> HttpRequest:
    body = get_normalised_body_from_serverless(event)

    headers = get_normalised_headers_from_serverless(event)
    headers["Content-Length"] = str(body.getbuffer().nbytes)

    raw_query_string = ""
    if event.get("multiValueQueryStringParameters"):
        for key, values in event.get("multiValueQueryStringParameters", {}).items():
            for value in values:
                raw_query_string += f"&{key}={quote_plus(value)}"

        raw_query_string = raw_query_string[1:]

    request = HttpRequest(
        method=event.get("httpMethod", "GET"),
        path=event.get("path", "/"),
        body=body,
        query_string=HttpQueryString(raw_query_string),
        headers=HttpHeaders(headers),
    )
    request.path_parameters = event.get("pathParameters", {})

    request.attributes["aws_context"] = context
    request.attributes["aws_event"] = event

    return request


def get_normalised_headers_from_serverless(event: Dict[str, Any]) -> Dict[str, str]:
    headers = event["headers"]
    request_context = event.get(
        "requestContext", {}
    )  # Set serverless related additional headers
    if request_context.get("requestId"):
        headers["x-serverless-request-id"] = request_context.get("requestId")
    if request_context.get("stage"):
        headers["x-serverless-stage"] = request_context.get("stage")
    if headers.get("x-amzn-trace-id"):
        headers["x-serverless-trace-id"] = headers["x-amzn-trace-id"]

    return headers


def get_normalised_body_from_serverless(event: Dict[str, Any]) -> BytesIO:
    body = event.get("body", "")
    if event.get("isBase64Encoded", False):
        body = base64.b64decode(body)

    if isinstance(body, str):
        body = body.encode("utf-8")

    return BytesIO(body)


def make_serverless_callback(
    middleware_pipeline: MiddlewarePipeline,
    func: Callable[[HttpRequest], HttpResponse],
    route: Route,
) -> Callable:
    def _handle_serverless_request(
        event: Dict[str, Any], context: Dict[str, Any]
    ) -> dict:

        if event.get("source") in [
            "aws.events",
            "serverless-plugin-warmup",
        ]:  # lambda warmup should be ignored
            return {
                "statusCode": int(HttpStatus.CONTINUE),
            }

        request = create_http_request_from_serverless_event(event, context)
        route._parameters = request.path_parameters
        request.route = route

        def response_middleware(
            _request: HttpRequest, _next: MiddlewareHandler
        ) -> HttpResponse:
            return func(_request)

        local_middleware = MiddlewarePipeline(middleware_pipeline.queue)
        local_middleware.append(response_middleware)

        try:
            response = local_middleware(request)
        except HttpError as http_error:
            response = HttpResponse(http_error.http_message, http_error.status_code)

        return make_serverless_response(event, response)

    return _handle_serverless_request


def make_serverless_response(
    event: Dict[str, Any], response: HttpResponse
) -> Dict[str, Any]:
    serverless_response: Dict[str, Any] = {"statusCode": int(response.status_code)}

    if "multiValueHeaders" in event:
        serverless_response["multiValueHeaders"] = response.headers._headers
    else:
        serverless_response["headers"] = {
            key: value for key, value in response.headers.items()
        }

    # If the request comes from ALB we need to add a status description
    is_elb = event.get("requestContext", {}).get("elb")
    if is_elb:
        serverless_response["statusDescription"] = str(response.status_code)

    mimetype = response.headers.get("content-type", "text/plain")
    if not isinstance(mimetype, str):
        mimetype = mimetype[0]

    body = str(response)

    if (
        mimetype.startswith("text/") or mimetype in TEXT_MIME_TYPES
    ) and not response.headers.get("Content-Encoding", ""):

        serverless_response["body"] = body
        serverless_response["isBase64Encoded"] = False
    else:
        serverless_response["body"] = base64.b64encode(body.encode("utf8"))
        serverless_response["isBase64Encoded"] = True

    return serverless_response



from cgi import parse_header
from copy import copy
from io import BytesIO
from typing import Any
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Union

from .http_cookies import HttpCookieJar
from .http_cookies import parse_cookie_header
from .http_headers import HttpHeaders
from .http_message import FormHttpMessage
from .http_message import HttpMessage
from .http_message import JsonHttpMessage
from .http_message import MultipartHttpMessage
from .http_method import HttpMethod
from .http_query_string import HttpQueryString
from .routing import Route


class HttpRequest:
    def __init__(
        self,
        method: Union[HttpMethod, str],
        path: str = "/",
        body: Union[Optional[BytesIO], str] = None,
        query_string: Union[Optional[HttpQueryString], str] = None,
        headers: Union[Optional[HttpHeaders], Dict[str, str]] = None,
    ):
        if isinstance(method, str):
            method = HttpMethod(method.upper())

        if isinstance(body, str):
            body = BytesIO(body.encode("utf8"))

        if isinstance(query_string, str):
            query_string = HttpQueryString(query_string)

        if isinstance(headers, dict):
            headers = HttpHeaders(headers)

        self.method = method
        self.path = path
        self.query_string = query_string
        self.path_parameters: Dict[str, str] = {}
        self.headers = headers if headers else HttpHeaders()
        self.route: Optional[Route] = None  # type: ignore
        self.attributes: Dict[str, Any] = {}
        self._body = body if body else BytesIO(b"")
        self._parsed_body: Optional[HttpMessage] = None
        self._cookies: Optional[HttpCookieJar] = None

    @property
    def body(self) -> BytesIO:
        return copy(self._body)

    @property
    def parsed_body(self) -> HttpMessage:
        if self._parsed_body:
            return copy(self._parsed_body)

        content_type: Tuple[str, Dict[str, str]] = parse_header(
            self.headers["Content-Type"]  # type: ignore
        )

        parsed_body: HttpMessage

        if content_type[0] == "multipart/form-data":
            parsed_body = MultipartHttpMessage.from_bytes(
                self.body,
                content_type[1].get("boundary", ""),
                content_type[1].get("charset", ""),
            )
        elif content_type[0] == "application/x-www-form-urlencoded":
            parsed_body = FormHttpMessage.from_bytes(
                self.body, content_type[1].get("charset", "utf8")
            )

        elif content_type[0] == "application/json":
            parsed_body = JsonHttpMessage.from_bytes(
                self.body, content_type[1].get("charset", "utf8")
            )
        else:
            self.body.seek(0)
            parsed_body = HttpMessage(
                self.body.read().decode(content_type[1].get("charset", "utf8"))
            )

        self._parsed_body = parsed_body

        return copy(self._parsed_body)

    @property
    def cookies(self):
        if self._cookies is None:
            self._cookies = parse_cookie_header(self.headers.get("cookie"))

        return copy(self._cookies)


__all__ = ["HttpRequest"]
